@@ -8,23 +8,57 @@ use nextest_filtering::errors::FiltersetParseErrors;
 use nextest_metadata::NextestExitCode;
 use nextest_runner::{
     config::core::{ConfigExperimental, ToolName},
-    errors::{UserConfigError, *},
+    errors::{
+        CacheDirError, RecordReadError, RecordSetupError, RunIdResolutionError, RunStoreError,
+        TestListFromSummaryError, UserConfigError, *,
+    },
     helpers::{format_interceptor_too_many_tests, plural},
     indenter::DisplayIndented,
     list::OwnedTestInstanceId,
+    record::{ReplayabilityStatus, RunListAlignment},
     redact::Redactor,
     run_mode::NextestRunMode,
     runner::{DebuggerCommand, TracerCommand},
 };
 use owo_colors::{OwoColorize, Style};
+use quick_junit::ReportUuid;
 use semver::Version;
-use std::{error::Error, io, path::PathBuf, process::ExitStatus, string::FromUtf8Error};
-use swrite::{SWrite, swriteln};
+use std::{error::Error, io, process::ExitStatus, string::FromUtf8Error};
+use swrite::{SWrite, swrite, swriteln};
 use thiserror::Error;
-use tracing::{Level, error, info};
+use tracing::{Level, error, info, warn};
 
 pub(crate) type Result<T, E = ExpectedError> = std::result::Result<T, E>;
 
+/// Error when combining incompatible cargo message format options.
+#[derive(Clone, Debug, PartialEq, Eq, Error)]
+pub enum CargoMessageFormatError {
+    /// Conflicting base formats specified.
+    #[error(
+        "conflicting message formats: `{first}` and `{second}` cannot be combined\n\
+         (only JSON modifiers like `json-diagnostic-short` can be combined)"
+    )]
+    ConflictingBaseFormats {
+        first: &'static str,
+        second: &'static str,
+    },
+
+    /// JSON modifier used with non-JSON base format.
+    #[error(
+        "cannot combine {modifiers} with `{base}` format\n\
+         (JSON modifiers can only be used with JSON output)"
+    )]
+    JsonModifierWithNonJson {
+        /// Comma-separated list of modifiers, each wrapped in backticks.
+        modifiers: &'static str,
+        base: &'static str,
+    },
+
+    /// Duplicate option specified.
+    #[error("duplicate message format option: `{option}`")]
+    Duplicate { option: &'static str },
+}
+
 #[derive(Debug)]
 #[doc(hidden)]
 pub enum ReuseBuildKind {
@@ -40,8 +74,6 @@ pub enum ReuseBuildKind {
 #[derive(Debug, Error)]
 #[doc(hidden)]
 pub enum ExpectedError {
-    #[error("could not change to requested directory")]
-    SetCurrentDirFailed { error: std::io::Error },
     #[error("failed to get current executable")]
     GetCurrentExeFailed {
         #[source]
@@ -222,7 +254,15 @@ pub enum ExpectedError {
     #[error("setup script failed")]
     SetupScriptFailed,
     #[error("test run failed")]
-    TestRunFailed,
+    TestRunFailed {
+        /// Whether the user can rerun failing tests with `-R latest`.
+        rerun_available: bool,
+    },
+    #[error("rerun tests outstanding")]
+    RerunTestsOutstanding {
+        /// The number of tests that were not seen during this rerun.
+        count: usize,
+    },
     #[error("no tests to run")]
     NoTestsRun {
         /// The run mode (test or benchmark).
@@ -288,6 +328,7 @@ pub enum ExpectedError {
         err: UpdateError,
     },
     #[error("error reading prompt")]
+    #[cfg(feature = "self-update")]
     DialoguerError {
         #[source]
         err: dialoguer::Error,
@@ -313,6 +354,21 @@ pub enum ExpectedError {
         config_file: Utf8PathBuf,
         missing: Vec<ConfigExperimental>,
     },
+    #[error("could not determine cache directory for recording")]
+    RecordCacheDirNotFound {
+        #[source]
+        err: CacheDirError,
+    },
+    #[error("error setting up recording")]
+    RecordSetupError {
+        #[source]
+        err: RunStoreError,
+    },
+    #[error("error setting up recording session")]
+    RecordSessionSetupError {
+        #[source]
+        err: RecordSetupError,
+    },
     #[error("filterset parse error")]
     FiltersetParseError {
         all_errors: Vec<FiltersetParseErrors>,
@@ -322,16 +378,18 @@ pub enum ExpectedError {
         reason: &'static str,
         args: Vec<String>,
     },
+    #[cfg(unix)]
     #[error("double-spawn parse error")]
     DoubleSpawnParseArgsError {
         args: String,
         #[source]
         err: shell_words::ParseError,
     },
+    #[cfg(unix)]
     #[error("double-spawn execution error")]
     DoubleSpawnExecError {
         command: Box<std::process::Command>,
-        current_dir: Result<PathBuf, std::io::Error>,
+        current_dir: Result<std::path::PathBuf, std::io::Error>,
         #[source]
         err: std::io::Error,
     },
@@ -340,6 +398,11 @@ pub enum ExpectedError {
         #[from]
         err: FormatVersionError,
     },
+    #[error("invalid cargo message format")]
+    CargoMessageFormatError {
+        #[from]
+        err: CargoMessageFormatError,
+    },
     #[error("extract read error")]
     DebugExtractReadError {
         kind: &'static str,
@@ -353,6 +416,34 @@ pub enum ExpectedError {
         #[source]
         err: std::io::Error,
     },
+    #[error("run ID resolution error")]
+    RunIdResolutionError {
+        #[source]
+        err: RunIdResolutionError,
+    },
+    #[error("error reading recorded run")]
+    RecordReadError {
+        #[source]
+        err: RecordReadError,
+    },
+    #[error(
+        "run {run_id} has unsupported store format version {found} (this nextest supports version {supported})"
+    )]
+    UnsupportedStoreFormatVersion {
+        run_id: ReportUuid,
+        found: u32,
+        supported: u32,
+    },
+    #[error("error reconstructing test list from archive")]
+    TestListFromSummaryError {
+        #[source]
+        err: TestListFromSummaryError,
+    },
+    #[error("write error")]
+    WriteError {
+        #[source]
+        err: std::io::Error,
+    },
 }
 
 impl ExpectedError {
@@ -457,8 +548,8 @@ impl ExpectedError {
         Self::SetupScriptFailed
     }
 
-    pub(crate) fn test_run_failed() -> Self {
-        Self::TestRunFailed
+    pub(crate) fn test_run_failed(rerun_available: bool) -> Self {
+        Self::TestRunFailed { rerun_available }
     }
 
     pub(crate) fn test_binary_args_parse_error(reason: &'static str, args: Vec<String>) -> Self {
@@ -474,7 +565,6 @@ impl ExpectedError {
             | Self::CargoLocateProjectFailed { .. } => NextestExitCode::CARGO_METADATA_FAILED,
             Self::WorkspaceRootInvalidUtf8 { .. }
             | Self::WorkspaceRootInvalid { .. }
-            | Self::SetCurrentDirFailed { .. }
             | Self::GetCurrentExeFailed { .. }
             | Self::ProfileNotFound { .. }
             | Self::StoreDirCreateError { .. }
@@ -494,11 +584,11 @@ impl ExpectedError {
             | Self::ConfigureHandleInheritanceError { .. }
             | Self::CargoMetadataParseError { .. }
             | Self::TestBinaryArgsParseError { .. }
-            | Self::DialoguerError { .. }
             | Self::SignalHandlerSetupError { .. }
             | Self::ShowTestGroupsError { .. }
             | Self::InvalidMessageFormatVersion { .. }
-            | Self::DebugExtractReadError { .. } => NextestExitCode::SETUP_ERROR,
+            | Self::DebugExtractReadError { .. }
+            | Self::CargoMessageFormatError { .. } => NextestExitCode::SETUP_ERROR,
             Self::ConfigParseError { err } => {
                 // Experimental features not being enabled are their own error.
                 match err.kind() {
@@ -510,7 +600,8 @@ impl ExpectedError {
             }
             Self::RequiredVersionNotMet { .. } => NextestExitCode::REQUIRED_VERSION_NOT_MET,
             #[cfg(feature = "self-update")]
-            Self::UpdateVersionParseError { .. } => NextestExitCode::SETUP_ERROR,
+            Self::DialoguerError { .. } | Self::UpdateVersionParseError { .. } => NextestExitCode::SETUP_ERROR,
+            #[cfg(unix)]
             Self::DoubleSpawnParseArgsError { .. } | Self::DoubleSpawnExecError { .. } => {
                 NextestExitCode::DOUBLE_SPAWN_ERROR
             }
@@ -521,7 +612,8 @@ impl ExpectedError {
                 NextestExitCode::BUILD_FAILED
             }
             Self::SetupScriptFailed => NextestExitCode::SETUP_SCRIPT_FAILED,
-            Self::TestRunFailed => NextestExitCode::TEST_RUN_FAILED,
+            Self::TestRunFailed { .. } => NextestExitCode::TEST_RUN_FAILED,
+            Self::RerunTestsOutstanding { .. } => NextestExitCode::RERUN_TESTS_OUTSTANDING,
             Self::NoTestsRun { .. } => NextestExitCode::NO_TESTS_RUN,
             Self::DebuggerNoTests { .. } | Self::DebuggerTooManyTests { .. }
             | Self::TracerNoTests { .. } | Self::TracerTooManyTests { .. } => NextestExitCode::SETUP_ERROR,
@@ -538,6 +630,14 @@ impl ExpectedError {
             | Self::ConfigExperimentalFeaturesNotEnabled { .. } => {
                 NextestExitCode::EXPERIMENTAL_FEATURE_NOT_ENABLED
             }
+            Self::RecordCacheDirNotFound { .. }
+            | Self::RecordSetupError { .. }
+            | Self::RecordSessionSetupError { .. }
+            | Self::RunIdResolutionError { .. }
+            | Self::RecordReadError { .. }
+            | Self::UnsupportedStoreFormatVersion { .. }
+            | Self::TestListFromSummaryError { .. } => NextestExitCode::SETUP_ERROR,
+            Self::WriteError { .. } => NextestExitCode::WRITE_OUTPUT_ERROR,
             Self::FiltersetParseError { .. } => NextestExitCode::INVALID_FILTERSET,
         }
     }
@@ -545,10 +645,6 @@ impl ExpectedError {
     /// Displays this error to stderr.
     pub fn display_to_stderr(&self, styles: &StderrStyles) {
         let mut next_error = match &self {
-            Self::SetCurrentDirFailed { error } => {
-                error!("could not change to requested directory");
-                Some(error as &dyn Error)
-            }
             Self::GetCurrentExeFailed { err } => {
                 error!("failed to get current executable");
                 Some(err as &dyn Error)
@@ -727,25 +823,27 @@ impl ExpectedError {
                                 "- unknown scripts specified (known scripts: {known_scripts_str}):",
                             );
                             for error in unknown_scripts {
-                                errors_str.push_str(&format!(
+                                swrite!(
+                                    errors_str,
                                     "  - script `{}` specified within profile `{}`\n",
                                     error.name.style(styles.bold),
                                     error.profile_name.style(styles.bold)
-                                ));
+                                );
                             }
                         }
 
                         if !wrong_script_types.is_empty() {
                             swriteln!(errors_str, "- scripts specified with incorrect type:");
                             for error in wrong_script_types {
-                                errors_str.push_str(&format!(
+                                swrite!(
+                                    errors_str,
                                     "  - script `{}` specified within profile `{}` as {}, \
                                      but is actually {}\n",
                                     error.name.style(styles.bold),
                                     error.profile_name.style(styles.bold),
                                     error.attempted.style(styles.bold),
                                     error.actual.style(styles.bold)
-                                ));
+                                );
                             }
                         }
 
@@ -786,7 +884,7 @@ impl ExpectedError {
                             .join(", ");
 
                         error!(
-                            "for config file `{}`{}, unknown experimental features defined:
+                            "for config file `{}`{}, unknown experimental features defined: \
                              {unknown_str} (known features: {known_str}):",
                             err.config_file(),
                             provided_by_tool(err.tool()),
@@ -938,8 +1036,33 @@ impl ExpectedError {
                 error!("setup script failed");
                 None
             }
-            Self::TestRunFailed => {
+            Self::TestRunFailed { rerun_available } => {
                 error!("test run failed");
+                if *rerun_available {
+                    info!(
+                        target: "cargo_nextest::no_heading",
+                        "(hint: {} to rerun failing tests)",
+                        "cargo nextest run -R latest".style(styles.bold),
+                    );
+                }
+                None
+            }
+            Self::RerunTestsOutstanding { count } => {
+                warn!(
+                    "{} outstanding {} still {}",
+                    count.style(styles.bold),
+                    // We only support reruns for test mode.
+                    plural::tests_str(NextestRunMode::Test, *count),
+                    plural::remains_str(*count),
+                );
+                // Advice to run `cargo nextest run -R latest` might not always
+                // be complete due to the expanded build scope or disappearing
+                // tests. We just hint that the user can continue doing reruns.
+                info!(
+                    target: "cargo_nextest::no_heading",
+                    "(hint: {} to continue rerunning)",
+                    "cargo nextest run -R latest".style(styles.bold),
+                );
                 None
             }
             Self::NoTestsRun { mode, is_default } => {
@@ -1046,6 +1169,7 @@ impl ExpectedError {
                 );
                 Some(err as &dyn Error)
             }
+            #[cfg(feature = "self-update")]
             Self::DialoguerError { err } => {
                 error!("error reading input prompt");
                 Some(err as &dyn Error)
@@ -1071,6 +1195,18 @@ impl ExpectedError {
                 );
                 None
             }
+            Self::RecordCacheDirNotFound { err } => {
+                error!("could not determine cache directory for recording test runs");
+                Some(err as &dyn Error)
+            }
+            Self::RecordSetupError { err } => {
+                error!("error setting up recording");
+                Some(err as &dyn Error)
+            }
+            Self::RecordSessionSetupError { err } => {
+                error!("error setting up recording session");
+                Some(err as &dyn Error)
+            }
             Self::FiltersetParseError { all_errors } => {
                 for errors in all_errors {
                     for single_error in &errors.errors {
@@ -1090,10 +1226,12 @@ impl ExpectedError {
                 );
                 None
             }
+            #[cfg(unix)]
             Self::DoubleSpawnParseArgsError { args, err } => {
                 error!("[double-spawn] failed to parse arguments `{args}`");
                 Some(err as &dyn Error)
             }
+            #[cfg(unix)]
             Self::DoubleSpawnExecError {
                 command,
                 current_dir,
@@ -1112,6 +1250,10 @@ impl ExpectedError {
                 error!("error parsing message format version");
                 Some(err as &dyn Error)
             }
+            Self::CargoMessageFormatError { err } => {
+                error!("invalid --cargo-message-format: {err}");
+                None
+            }
             Self::DebugExtractReadError { kind, path, err } => {
                 error!("error reading {kind} file `{}`", path.style(styles.bold),);
                 Some(err as &dyn Error)
@@ -1120,6 +1262,87 @@ impl ExpectedError {
                 error!("error writing {format} output");
                 Some(err as &dyn Error)
             }
+            Self::RunIdResolutionError { err } => {
+                match &err {
+                    RunIdResolutionError::NotFound { prefix } => {
+                        error!(
+                            "no recorded run found matching `{}`",
+                            prefix.style(styles.bold)
+                        );
+                    }
+                    RunIdResolutionError::Ambiguous {
+                        prefix,
+                        count,
+                        candidates,
+                        run_id_index,
+                    } => {
+                        error!(
+                            "prefix `{}` is ambiguous, matches {} {}:",
+                            prefix.style(styles.bold),
+                            count.style(styles.bold),
+                            plural::runs_str(*count)
+                        );
+                        let redactor = if crate::output::should_redact() {
+                            Redactor::for_snapshot_testing()
+                        } else {
+                            Redactor::noop()
+                        };
+                        let alignment = RunListAlignment::from_runs(candidates);
+                        for candidate in candidates {
+                            info!(
+                                target: "cargo_nextest::no_heading",
+                                "{}",
+                                candidate.display(run_id_index, &ReplayabilityStatus::Replayable, alignment, &styles.record_styles, &redactor)
+                            );
+                        }
+                        if *count > candidates.len() {
+                            let remaining = *count - candidates.len();
+                            info!(
+                                target: "cargo_nextest::no_heading",
+                                "  ... and {} more {}",
+                                remaining.style(styles.bold),
+                                plural::runs_str(remaining)
+                            );
+                        }
+                    }
+                    RunIdResolutionError::InvalidPrefix { prefix } => {
+                        error!(
+                            "prefix `{}` contains invalid characters (expected hexadecimal)",
+                            prefix.style(styles.bold)
+                        );
+                    }
+                    RunIdResolutionError::NoRuns => {
+                        error!("no recorded runs exist");
+                    }
+                }
+                None
+            }
+            Self::RecordReadError { err } => {
+                error!("error reading recorded run");
+                Some(err as &dyn Error)
+            }
+            Self::UnsupportedStoreFormatVersion {
+                run_id,
+                found,
+                supported,
+            } => {
+                error!(
+                    "run {} has unsupported store format version {} \
+                     (this nextest supports version {})",
+                    run_id.style(styles.bold),
+                    found.style(styles.bold),
+                    supported.style(styles.bold),
+                );
+                None
+            }
+            Self::TestListFromSummaryError { err } => {
+                error!("error reconstructing test list from archived summary");
+                Some(err as &dyn Error)
+            }
+            Self::WriteError { err } => {
+                error!("write error");
+                Some(err as &dyn Error)
+            }
         };
 
         while let Some(err) = next_error {
