@@ -28,6 +28,7 @@ use indexmap::IndexMap;
 use nextest_filtering::{
     BinaryQuery, EvalContext, Filterset, FiltersetKind, ParseContext, TestQuery,
 };
+use quick_junit::ReportUuid;
 use serde::{Deserialize, de::Error};
 use smol_str::SmolStr;
 use std::{
@@ -36,6 +37,7 @@ use std::{
     process::Command,
     sync::Arc,
 };
+use swrite::{SWrite, swrite};
 
 /// The scripts defined in nextest configuration.
 #[derive(Clone, Debug, Default, Deserialize)]
@@ -533,7 +535,8 @@ impl CompiledProfileScripts<FinalConfig> {
 }
 
 /// The name of a configuration script.
-#[derive(Clone, Debug, Eq, PartialEq, Hash, PartialOrd, Ord)]
+#[derive(Clone, Debug, Eq, PartialEq, Hash, PartialOrd, Ord, serde::Serialize)]
+#[serde(transparent)]
 pub struct ScriptId(pub ConfigIdentifier);
 
 impl ScriptId {
@@ -548,6 +551,16 @@ impl ScriptId {
         &self.0
     }
 
+    /// Returns a unique ID for this script, consisting of the run ID, the script ID, and the stress index.
+    pub fn unique_id(&self, run_id: ReportUuid, stress_index: Option<u32>) -> String {
+        let mut out = String::new();
+        swrite!(out, "{run_id}:{self}");
+        if let Some(stress_index) = stress_index {
+            swrite!(out, "@stress-{}", stress_index);
+        }
+        out
+    }
+
     #[cfg(test)]
     pub(super) fn as_str(&self) -> &str {
         self.0.as_str()
@@ -1008,7 +1021,7 @@ mod tests {
     use super::*;
     use crate::{
         config::{
-            core::{ConfigExperimental, NextestConfig, ToolConfigFile},
+            core::{ConfigExperimental, NextestConfig, ToolConfigFile, ToolName},
             utils::test_helpers::*,
         },
         errors::{
@@ -1020,8 +1033,13 @@ mod tests {
     use camino_tempfile_ext::prelude::*;
     use indoc::indoc;
     use maplit::btreeset;
+    use nextest_metadata::TestCaseName;
     use test_case::test_case;
 
+    fn tool_name(s: &str) -> ToolName {
+        ToolName::new(s.into()).unwrap()
+    }
+
     #[test]
     fn test_scripts_basic() {
         let config_contents = indoc! {r#"
@@ -1074,7 +1092,7 @@ mod tests {
         let pcx = ParseContext::new(&graph);
 
         let tool_config_files = [ToolConfigFile {
-            tool: "my-tool".to_owned(),
+            tool: tool_name("my-tool"),
             config_file: tool_path.to_path_buf(),
         }];
 
@@ -1114,9 +1132,10 @@ mod tests {
         // This query matches the foo and bar scripts.
         let host_binary_query =
             binary_query(&graph, package_id, "lib", "my-binary", BuildPlatform::Host);
+        let test_name = TestCaseName::new("script1");
         let query = TestQuery {
             binary_query: host_binary_query.to_query(),
-            test_name: "script1",
+            test_name: &test_name,
         };
         let scripts = SetupScripts::new_with_queries(&profile, std::iter::once(query));
         assert_eq!(scripts.len(), 2, "two scripts should be enabled");
@@ -1140,9 +1159,10 @@ mod tests {
         );
 
         // This query matches the baz script.
+        let test_name = TestCaseName::new("script2");
         let query = TestQuery {
             binary_query: target_binary_query.to_query(),
-            test_name: "script2",
+            test_name: &test_name,
         };
         let scripts = SetupScripts::new_with_queries(&profile, std::iter::once(query));
         assert_eq!(scripts.len(), 1, "one script should be enabled");
@@ -1153,9 +1173,10 @@ mod tests {
         );
 
         // This query matches the baz, foo and tool scripts (but note the order).
+        let test_name = TestCaseName::new("script3");
         let query = TestQuery {
             binary_query: target_binary_query.to_query(),
-            test_name: "script3",
+            test_name: &test_name,
         };
         let scripts = SetupScripts::new_with_queries(&profile, std::iter::once(query));
         assert_eq!(scripts.len(), 3, "three scripts should be enabled");
@@ -1514,7 +1535,7 @@ mod tests {
         let tool_path = workspace_dir.child(".config/my-tool.toml");
         tool_path.write_str(tool_config_contents).unwrap();
         let tool_config_files = [ToolConfigFile {
-            tool: "my-tool".to_owned(),
+            tool: tool_name("my-tool"),
             config_file: tool_path.to_path_buf(),
         }];
 
