@@ -6,27 +6,31 @@
 use crate::{
     cargo_config::{TargetTriple, TargetTripleSource},
     config::{
-        core::ConfigExperimental,
+        core::{ConfigExperimental, ToolName},
         elements::{CustomTestGroup, TestGroup},
         scripts::{ProfileScriptType, ScriptId, ScriptType},
     },
     helpers::{display_exited_with, dylib_path_envvar},
+    indenter::{DisplayIndented, indented},
+    record::{RecordedRunInfo, RunIdIndex},
     redact::Redactor,
     reuse_build::{ArchiveFormat, ArchiveStep},
     target_runner::PlatformRunnerSource,
 };
 use camino::{FromPathBufError, Utf8Path, Utf8PathBuf};
 use config::ConfigError;
-use indent_write::{fmt::IndentWriter, indentable::Indented};
 use itertools::{Either, Itertools};
 use nextest_filtering::errors::FiltersetParseErrors;
 use nextest_metadata::RustBinaryId;
+use quick_junit::ReportUuid;
+use serde::{Deserialize, Serialize};
 use smol_str::SmolStr;
 use std::{
     borrow::Cow,
     collections::BTreeSet,
     env::JoinPathsError,
     fmt::{self, Write as _},
+    path::PathBuf,
     process::ExitStatus,
     sync::Arc,
 };
@@ -37,12 +41,12 @@ use thiserror::Error;
 #[derive(Debug, Error)]
 #[error(
     "failed to parse nextest config at `{config_file}`{}",
-    provided_by_tool(tool.as_deref())
+    provided_by_tool(tool.as_ref())
 )]
 #[non_exhaustive]
 pub struct ConfigParseError {
     config_file: Utf8PathBuf,
-    tool: Option<String>,
+    tool: Option<ToolName>,
     #[source]
     kind: ConfigParseErrorKind,
 }
@@ -50,12 +54,12 @@ pub struct ConfigParseError {
 impl ConfigParseError {
     pub(crate) fn new(
         config_file: impl Into<Utf8PathBuf>,
-        tool: Option<&str>,
+        tool: Option<&ToolName>,
         kind: ConfigParseErrorKind,
     ) -> Self {
         Self {
             config_file: config_file.into(),
-            tool: tool.map(|s| s.to_owned()),
+            tool: tool.cloned(),
             kind,
         }
     }
@@ -66,8 +70,8 @@ impl ConfigParseError {
     }
 
     /// Returns the tool name associated with this error.
-    pub fn tool(&self) -> Option<&str> {
-        self.tool.as_deref()
+    pub fn tool(&self) -> Option<&ToolName> {
+        self.tool.as_ref()
     }
 
     /// Returns the kind of error this is.
@@ -77,7 +81,7 @@ impl ConfigParseError {
 }
 
 /// Returns the string ` provided by tool <tool>`, if `tool` is `Some`.
-pub fn provided_by_tool(tool: Option<&str>) -> String {
+pub fn provided_by_tool(tool: Option<&ToolName>) -> String {
     match tool {
         Some(tool) => format!(" provided by tool `{tool}`"),
         None => String::new(),
@@ -93,6 +97,9 @@ pub enum ConfigParseErrorKind {
     /// An error occurred while building the config.
     #[error(transparent)]
     BuildError(Box<ConfigError>),
+    /// An error occurred while parsing the config into a table.
+    #[error(transparent)]
+    TomlParseError(Box<toml::de::Error>),
     #[error(transparent)]
     /// An error occurred while deserializing the config.
     DeserializeError(Box<serde_path_to_error::Error<ConfigError>>),
@@ -178,6 +185,9 @@ pub enum ConfigParseErrorKind {
         /// The features that were not enabled.
         missing_features: BTreeSet<ConfigExperimental>,
     },
+    /// An inheritance cycle was detected in the profile configuration.
+    #[error("inheritance error(s) detected: {}", .0.iter().join(", "))]
+    InheritanceErrors(Vec<InheritsError>),
 }
 
 /// An error that occurred while compiling overrides or scripts specified in
@@ -355,10 +365,10 @@ pub enum SetupScriptOutputError {
 ///
 /// In the future, we'll likely want to replace this with a `miette::Diagnostic`-based error, since
 /// that supports multiple causes via "related".
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
 pub struct ErrorList<T> {
     // A description of what the errors are.
-    description: &'static str,
+    description: Cow<'static, str>,
     // Invariant: this list is non-empty.
     inner: Vec<T>,
 }
@@ -372,7 +382,7 @@ impl<T: std::error::Error> ErrorList<T> {
             None
         } else {
             Some(Self {
-                description,
+                description: Cow::Borrowed(description),
                 inner: errors.into_iter().map(T::from).collect(),
             })
         }
@@ -388,9 +398,36 @@ impl<T: std::error::Error> ErrorList<T> {
         }
     }
 
-    pub(crate) fn iter(&self) -> impl Iterator<Item = &T> {
+    /// Returns the description of what the errors are.
+    pub fn description(&self) -> &str {
+        &self.description
+    }
+
+    /// Iterates over the errors in this list.
+    pub fn iter(&self) -> impl Iterator<Item = &T> {
         self.inner.iter()
     }
+
+    /// Transforms the errors in this list using the given function.
+    pub fn map<U, F>(self, f: F) -> ErrorList<U>
+    where
+        U: std::error::Error,
+        F: FnMut(T) -> U,
+    {
+        ErrorList {
+            description: self.description,
+            inner: self.inner.into_iter().map(f).collect(),
+        }
+    }
+}
+
+impl<T: std::error::Error> IntoIterator for ErrorList<T> {
+    type Item = T;
+    type IntoIter = std::vec::IntoIter<T>;
+
+    fn into_iter(self) -> Self::IntoIter {
+        self.inner.into_iter()
+    }
 }
 
 impl<T: std::error::Error> fmt::Display for ErrorList<T> {
@@ -408,7 +445,7 @@ impl<T: std::error::Error> fmt::Display for ErrorList<T> {
             self.description,
         )?;
         for error in &self.inner {
-            let mut indent = IndentWriter::new_skip_initial("  ", f);
+            let mut indent = indented(f).with_str("  ").skip_initial();
             writeln!(indent, "* {}", DisplayErrorChain::new(error))?;
             f = indent.into_inner();
         }
@@ -416,6 +453,26 @@ impl<T: std::error::Error> fmt::Display for ErrorList<T> {
     }
 }
 
+#[cfg(test)]
+impl<T: proptest::arbitrary::Arbitrary + std::fmt::Debug + 'static> proptest::arbitrary::Arbitrary
+    for ErrorList<T>
+{
+    type Parameters = ();
+    type Strategy = proptest::strategy::BoxedStrategy<Self>;
+
+    fn arbitrary_with(_: Self::Parameters) -> Self::Strategy {
+        use proptest::prelude::*;
+
+        // Generate 1-5 errors (non-empty).
+        proptest::collection::vec(any::<T>(), 1..=5)
+            .prop_map(|inner| ErrorList {
+                description: Cow::Borrowed("test errors"),
+                inner,
+            })
+            .boxed()
+    }
+}
+
 impl<T: std::error::Error> std::error::Error for ErrorList<T> {
     fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
         if self.inner.len() == 1 {
@@ -430,22 +487,24 @@ impl<T: std::error::Error> std::error::Error for ErrorList<T> {
 
 /// A wrapper type to display a chain of errors with internal indentation.
 ///
-/// This is similar to the display-error-chain crate, but uses IndentWriter
+/// This is similar to the display-error-chain crate, but uses an indenter
 /// internally to ensure that subsequent lines are also nested.
-pub(crate) struct DisplayErrorChain<E> {
+pub struct DisplayErrorChain<E> {
     error: E,
     initial_indent: &'static str,
 }
 
 impl<E: std::error::Error> DisplayErrorChain<E> {
-    pub(crate) fn new(error: E) -> Self {
+    /// Creates a new `DisplayErrorChain` with the given error.
+    pub fn new(error: E) -> Self {
         Self {
             error,
             initial_indent: "",
         }
     }
 
-    pub(crate) fn new_with_initial_indent(initial_indent: &'static str, error: E) -> Self {
+    /// Creates a new `DisplayErrorChain` with the given error and initial indentation.
+    pub fn new_with_initial_indent(initial_indent: &'static str, error: E) -> Self {
         Self {
             error,
             initial_indent,
@@ -458,7 +517,7 @@ where
     E: std::error::Error,
 {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        let mut writer = IndentWriter::new(self.initial_indent, f);
+        let mut writer = indented(f).with_str(self.initial_indent);
         write!(writer, "{}", self.error)?;
 
         let Some(mut cause) = self.error.source() else {
@@ -469,7 +528,8 @@ where
 
         loop {
             writeln!(writer)?;
-            let mut indent = IndentWriter::new_skip_initial("    ", writer);
+            // Wrap the existing writer to accumulate indentation.
+            let mut indent = indented(&mut writer).with_str("    ").skip_initial();
             write!(indent, "  - {cause}")?;
 
             let Some(next_cause) = cause.source() else {
@@ -477,7 +537,6 @@ where
             };
 
             cause = next_cause;
-            writer = indent.into_inner();
         }
     }
 }
@@ -594,7 +653,7 @@ impl ProfileScriptErrors {
 
 /// An error which indicates that a profile was requested but not known to nextest.
 #[derive(Clone, Debug, Error)]
-#[error("profile `{profile} not found (known profiles: {})`", .all_profiles.join(", "))]
+#[error("profile `{profile}` not found (known profiles: {})", .all_profiles.join(", "))]
 pub struct ProfileNotFound {
     profile: String,
     all_profiles: Vec<String>,
@@ -638,6 +697,22 @@ pub enum InvalidIdentifier {
     ToolIdentifierInvalidXid(SmolStr),
 }
 
+/// A tool name is invalid.
+#[derive(Clone, Debug, Error, Eq, PartialEq)]
+pub enum InvalidToolName {
+    /// The tool name is empty.
+    #[error("tool name is empty")]
+    Empty,
+
+    /// The tool name is not in the correct Unicode format.
+    #[error("invalid tool name `{0}`")]
+    InvalidXid(SmolStr),
+
+    /// The tool name starts with "@tool", which is reserved for tool identifiers.
+    #[error("tool name cannot start with \"@tool\": `{0}`")]
+    StartsWithToolPrefix(SmolStr),
+}
+
 /// The name of a test group is invalid (not a valid identifier).
 #[derive(Clone, Debug, Error)]
 #[error("invalid custom test group name: {0}")]
@@ -649,7 +724,7 @@ pub struct InvalidCustomTestGroupName(pub InvalidIdentifier);
 pub struct InvalidConfigScriptName(pub InvalidIdentifier);
 
 /// Error returned while parsing a [`ToolConfigFile`](crate::config::core::ToolConfigFile) value.
-#[derive(Clone, Debug, Error)]
+#[derive(Clone, Debug, Error, PartialEq, Eq)]
 pub enum ToolConfigFileParseError {
     #[error(
         "tool-config-file has invalid format: {input}\n(hint: tool configs must be in the format <tool-name>:<path>)"
@@ -660,11 +735,15 @@ pub enum ToolConfigFileParseError {
         input: String,
     },
 
-    /// The tool name was empty.
-    #[error("tool-config-file has empty tool name: {input}")]
-    EmptyToolName {
+    /// The tool name was invalid.
+    #[error("tool-config-file has invalid tool name: {input}")]
+    InvalidToolName {
         /// The input that failed to parse.
         input: String,
+
+        /// The error that occurred.
+        #[source]
+        error: InvalidToolName,
     },
 
     /// The config file path was empty.
@@ -682,6 +761,61 @@ pub enum ToolConfigFileParseError {
     },
 }
 
+/// Errors that can occur while loading user config.
+#[derive(Debug, Error)]
+#[non_exhaustive]
+pub enum UserConfigError {
+    /// The user config file specified via `--user-config-file` or
+    /// `NEXTEST_USER_CONFIG_FILE` does not exist.
+    #[error("user config file not found at {path}")]
+    FileNotFound {
+        /// The path that was specified.
+        path: Utf8PathBuf,
+    },
+
+    /// Failed to read the user config file.
+    #[error("failed to read user config at {path}")]
+    Read {
+        /// The path to the config file.
+        path: Utf8PathBuf,
+        /// The underlying I/O error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// Failed to parse the user config file.
+    #[error("failed to parse user config at {path}")]
+    Parse {
+        /// The path to the config file.
+        path: Utf8PathBuf,
+        /// The underlying TOML parse error.
+        #[source]
+        error: toml::de::Error,
+    },
+
+    /// The user config path contains non-UTF-8 characters.
+    #[error("user config path contains non-UTF-8 characters")]
+    NonUtf8Path {
+        /// The underlying error from path conversion.
+        #[source]
+        error: FromPathBufError,
+    },
+
+    /// Failed to compile a platform spec in an override.
+    #[error(
+        "for user config at {path}, failed to compile platform spec in [[overrides]] at index {index}"
+    )]
+    OverridePlatformSpec {
+        /// The path to the config file.
+        path: Utf8PathBuf,
+        /// The index of the override in the array.
+        index: usize,
+        /// The underlying target-spec error.
+        #[source]
+        error: target_spec::Error,
+    },
+}
+
 /// Error returned while parsing a [`MaxFail`](crate::config::elements::MaxFail) input.
 #[derive(Clone, Debug, Error)]
 #[error("unrecognized value for max-fail: {reason}")]
@@ -730,6 +864,19 @@ pub enum DebuggerCommandParseError {
     EmptyCommand,
 }
 
+/// An error that occurred while parsing a tracer command.
+#[derive(Clone, Debug, Error)]
+#[non_exhaustive]
+pub enum TracerCommandParseError {
+    /// The command string could not be parsed as shell words.
+    #[error(transparent)]
+    ShellWordsParse(shell_words::ParseError),
+
+    /// The command was empty.
+    #[error("tracer command cannot be empty")]
+    EmptyCommand,
+}
+
 /// Error returned while parsing a [`TestThreads`](crate::config::elements::TestThreads) value.
 #[derive(Clone, Debug, Error)]
 #[error(
@@ -1617,8 +1764,8 @@ pub enum HostPlatformDetectError {
          - `rustc -vV` stderr:\n{}\n\
          - build target error:\n{}\n",
         status,
-        Indented { item: String::from_utf8_lossy(stdout), indent: "  " },
-        Indented { item: String::from_utf8_lossy(stderr), indent: "  " },
+        DisplayIndented { item: String::from_utf8_lossy(stdout), indent: "  " },
+        DisplayIndented { item: String::from_utf8_lossy(stderr), indent: "  " },
         DisplayErrorChain::new_with_initial_indent("  ", build_target_error)
     )]
     RustcVvFailed {
@@ -1809,16 +1956,632 @@ pub enum ShowTestGroupsError {
     },
 }
 
+/// An error occurred while processing profile's inherits setting
+#[derive(Debug, Error, PartialEq, Eq, Hash)]
+pub enum InheritsError {
+    /// The default profile should not be able to inherit from other profiles
+    #[error("the {} profile should not inherit from other profiles", .0)]
+    DefaultProfileInheritance(String),
+    /// An unknown/unfound profile was detected to inherit from in profile configuration
+    #[error("profile {} inherits from an unknown profile {}", .0, .1)]
+    UnknownInheritance(String, String),
+    /// A self referential inheritance is detected from this profile
+    #[error("a self referential inheritance is detected from profile: {}", .0)]
+    SelfReferentialInheritance(String),
+    /// An inheritance cycle was detected in the profile configuration.
+    #[error("inheritance cycle detected in profile configuration from: {}", .0.iter().map(|scc| {
+        format!("[{}]", scc.iter().join(", "))
+    }).join(", "))]
+    InheritanceCycle(Vec<Vec<String>>),
+}
+
+// ---
+// Record and replay errors
+// ---
+
+/// An error that occurred while managing the run store.
+#[derive(Debug, Error)]
+pub enum RunStoreError {
+    /// An error occurred while creating the run directory.
+    #[error("error creating run directory `{run_dir}`")]
+    RunDirCreate {
+        /// The run directory that could not be created.
+        run_dir: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while acquiring a file lock.
+    #[error("error acquiring lock on `{path}`")]
+    FileLock {
+        /// The path to the lock file.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// Timed out waiting to acquire a file lock.
+    #[error(
+        "timed out acquiring lock on `{path}` after {timeout_secs}s (is the cache directory \
+         on a networked filesystem?)"
+    )]
+    FileLockTimeout {
+        /// The path to the lock file.
+        path: Utf8PathBuf,
+
+        /// The timeout duration in seconds.
+        timeout_secs: u64,
+    },
+
+    /// An error occurred while reading the run list.
+    #[error("error reading run list from `{path}`")]
+    RunListRead {
+        /// The path to the run list file.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while deserializing the run list.
+    #[error("error deserializing run list from `{path}`")]
+    RunListDeserialize {
+        /// The path to the run list file.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: serde_json::Error,
+    },
+
+    /// An error occurred while serializing the run list.
+    #[error("error serializing run list to `{path}`")]
+    RunListSerialize {
+        /// The path to the run list file.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: serde_json::Error,
+    },
+
+    /// An error occurred while serializing rerun info.
+    #[error("error serializing rerun info")]
+    RerunInfoSerialize {
+        /// The underlying error.
+        #[source]
+        error: serde_json::Error,
+    },
+
+    /// An error occurred while serializing the test list.
+    #[error("error serializing test list")]
+    TestListSerialize {
+        /// The underlying error.
+        #[source]
+        error: serde_json::Error,
+    },
+
+    /// An error occurred while serializing the record options.
+    #[error("error serializing record options")]
+    RecordOptionsSerialize {
+        /// The underlying error.
+        #[source]
+        error: serde_json::Error,
+    },
+
+    /// An error occurred while serializing a test event.
+    #[error("error serializing test event")]
+    TestEventSerialize {
+        /// The underlying error.
+        #[source]
+        error: serde_json::Error,
+    },
+
+    /// An error occurred while writing the run list.
+    #[error("error writing run list to `{path}`")]
+    RunListWrite {
+        /// The path to the run list file.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: atomicwrites::Error<std::io::Error>,
+    },
+
+    /// An error occurred while writing to the store.
+    #[error("error writing to store at `{store_path}`")]
+    StoreWrite {
+        /// The path to the store file.
+        store_path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: StoreWriterError,
+    },
+
+    /// An error occurred while creating the run log.
+    #[error("error creating run log at `{path}`")]
+    RunLogCreate {
+        /// The path to the run log file.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while writing to the run log.
+    #[error("error writing to run log at `{path}`")]
+    RunLogWrite {
+        /// The path to the run log file.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while flushing the run log.
+    #[error("error flushing run log at `{path}`")]
+    RunLogFlush {
+        /// The path to the run log file.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// Cannot write to runs.json.zst because it has a newer format version.
+    #[error(
+        "cannot write to record store: runs.json.zst format version {file_version} is newer than \
+         supported version {max_supported_version}"
+    )]
+    FormatVersionTooNew {
+        /// The format version in the file.
+        file_version: u32,
+        /// The maximum version this nextest can write.
+        max_supported_version: u32,
+    },
+}
+
+/// An error that occurred while writing to a zip store.
+#[derive(Debug, Error)]
+#[non_exhaustive]
+pub enum StoreWriterError {
+    /// An error occurred while creating the store file.
+    #[error("error creating store")]
+    Create {
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while starting a new file in the store.
+    #[error("error creating path `{path}` in store")]
+    StartFile {
+        /// The path within the store.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: zip::result::ZipError,
+    },
+
+    /// An error occurred while writing to a file in the store.
+    #[error("error writing to path `{path}` in store")]
+    Write {
+        /// The path within the store.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while compressing data with a zstd dictionary.
+    #[error("error compressing data")]
+    Compress {
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while finalizing the store.
+    #[error("error finalizing store")]
+    Finish {
+        /// The underlying error.
+        #[source]
+        error: zip::result::ZipError,
+    },
+
+    /// An error occurred while flushing the store.
+    #[error("error flushing store")]
+    Flush {
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+}
+
+/// An error that occurred in the record reporter.
+#[derive(Debug, Error)]
+pub enum RecordReporterError {
+    /// An error occurred while writing to the run store.
+    #[error(transparent)]
+    RunStore(RunStoreError),
+
+    /// The writer thread panicked.
+    #[error("record writer thread panicked: {message}")]
+    WriterPanic {
+        /// A message extracted from the panic payload, if available.
+        message: String,
+    },
+}
+
+/// An error determining the cache directory for recordings.
+#[derive(Debug, Error)]
+pub enum CacheDirError {
+    /// The platform base strategy could not be determined.
+    ///
+    /// This typically means the platform doesn't support standard directory layouts.
+    #[error("could not determine platform base directory strategy")]
+    BaseDirStrategy,
+
+    /// The platform cache directory path is not valid UTF-8.
+    #[error("platform cache directory is not valid UTF-8: {path:?}")]
+    CacheDirNotUtf8 {
+        /// The path that was not valid UTF-8.
+        path: PathBuf,
+    },
+
+    /// The workspace path could not be canonicalized.
+    #[error("could not canonicalize workspace path `{workspace_root}`")]
+    Canonicalize {
+        /// The workspace root that could not be canonicalized.
+        workspace_root: Utf8PathBuf,
+        /// The underlying I/O error.
+        #[source]
+        error: std::io::Error,
+    },
+}
+
+/// An error during recording session setup.
+#[derive(Debug, Error)]
+pub enum RecordSetupError {
+    /// The platform cache directory could not be determined.
+    #[error("could not determine platform cache directory for recording")]
+    CacheDirNotFound(#[source] CacheDirError),
+
+    /// Failed to create the run store.
+    #[error("failed to create run store")]
+    StoreCreate(#[source] RunStoreError),
+
+    /// Failed to acquire exclusive lock on the run store.
+    #[error("failed to lock run store")]
+    StoreLock(#[source] RunStoreError),
+
+    /// Failed to create the run recorder.
+    #[error("failed to create run recorder")]
+    RecorderCreate(#[source] RunStoreError),
+}
+
+impl RecordSetupError {
+    /// If this error indicates that recording should be disabled (but the test
+    /// run should continue), returns the underlying error.
+    ///
+    /// This is the case when the runs.json.zst file has a newer format version than
+    /// this nextest supports. Recording is disabled to avoid data loss, but the
+    /// test run can proceed normally.
+    pub fn disabled_error(&self) -> Option<&(dyn std::error::Error + 'static)> {
+        match self {
+            RecordSetupError::RecorderCreate(err @ RunStoreError::FormatVersionTooNew { .. }) => {
+                Some(err)
+            }
+            _ => None,
+        }
+    }
+}
+
+/// An error that occurred while pruning recorded runs.
+#[derive(Debug, Error)]
+pub enum RecordPruneError {
+    /// An error occurred while deleting a run directory.
+    #[error("error deleting run `{run_id}` at `{path}`")]
+    DeleteRun {
+        /// The run ID that could not be deleted.
+        run_id: ReportUuid,
+
+        /// The path to the run directory.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while calculating the size of a path.
+    #[error("error calculating size of `{path}`")]
+    CalculateSize {
+        /// The path whose size could not be calculated.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while deleting an orphaned directory.
+    #[error("error deleting orphaned directory `{path}`")]
+    DeleteOrphan {
+        /// The path to the orphaned directory.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while reading the runs directory.
+    #[error("error reading runs directory `{path}`")]
+    ReadRunsDir {
+        /// The path to the runs directory.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while reading a directory entry.
+    #[error("error reading directory entry in `{dir}`")]
+    ReadDirEntry {
+        /// The path to the directory being read.
+        dir: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// An error occurred while reading file type.
+    #[error("error reading file type for `{path}`")]
+    ReadFileType {
+        /// The path whose file type could not be read.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+}
+
+/// Error returned when parsing an invalid run ID selector.
+///
+/// A valid selector is either "latest" or a string containing only hex digits
+/// and dashes (for UUID format).
+#[derive(Clone, Debug, PartialEq, Eq, Error)]
+#[error("invalid run ID selector `{input}`: expected `latest` or hex digits")]
+pub struct InvalidRunIdSelector {
+    /// The invalid input string.
+    pub input: String,
+}
+
+/// An error resolving a run ID prefix.
+#[derive(Debug, Error)]
+pub enum RunIdResolutionError {
+    /// No run found matching the prefix.
+    #[error("no recorded run found matching `{prefix}`")]
+    NotFound {
+        /// The prefix that was searched for.
+        prefix: String,
+    },
+
+    /// Multiple runs match the prefix.
+    #[error("prefix `{prefix}` is ambiguous, matches {count} runs")]
+    Ambiguous {
+        /// The prefix that was searched for.
+        prefix: String,
+
+        /// The number of matching runs.
+        count: usize,
+
+        /// The candidates that matched (up to a limit).
+        candidates: Vec<RecordedRunInfo>,
+
+        /// The run ID index for computing shortest unique prefixes.
+        run_id_index: RunIdIndex,
+    },
+
+    /// The prefix contains invalid characters.
+    #[error("prefix `{prefix}` contains invalid characters (expected hexadecimal)")]
+    InvalidPrefix {
+        /// The invalid prefix.
+        prefix: String,
+    },
+
+    /// No recorded runs exist.
+    #[error("no recorded runs exist")]
+    NoRuns,
+}
+
+/// An error that occurred while reading a recorded run.
+#[derive(Debug, Error)]
+pub enum RecordReadError {
+    /// The run was not found.
+    #[error("run not found at `{path}`")]
+    RunNotFound {
+        /// The path where the run was expected.
+        path: Utf8PathBuf,
+    },
+
+    /// Failed to open the archive.
+    #[error("error opening archive at `{path}`")]
+    OpenArchive {
+        /// The path to the archive.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// Failed to read an archive file.
+    #[error("error reading `{file_name}` from archive")]
+    ReadArchiveFile {
+        /// The file name within the archive.
+        file_name: String,
+
+        /// The underlying error.
+        #[source]
+        error: zip::result::ZipError,
+    },
+
+    /// Failed to open the run log.
+    #[error("error opening run log at `{path}`")]
+    OpenRunLog {
+        /// The path to the run log.
+        path: Utf8PathBuf,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// Failed to read a line from the run log.
+    #[error("error reading line {line_number} from run log")]
+    ReadRunLog {
+        /// The line number that failed.
+        line_number: usize,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// Failed to parse an event from the run log.
+    #[error("error parsing event at line {line_number}")]
+    ParseEvent {
+        /// The line number that failed.
+        line_number: usize,
+
+        /// The underlying error.
+        #[source]
+        error: serde_json::Error,
+    },
+
+    /// Required file not found in archive.
+    #[error("required file `{file_name}` not found in archive")]
+    FileNotFound {
+        /// The file name that was not found.
+        file_name: String,
+    },
+
+    /// Failed to decompress archive data.
+    #[error("error decompressing data from `{file_name}`")]
+    Decompress {
+        /// The file name being decompressed.
+        file_name: String,
+
+        /// The underlying error.
+        #[source]
+        error: std::io::Error,
+    },
+
+    /// Unknown output file type in archive.
+    ///
+    /// This indicates the archive was created by a newer version of nextest
+    /// with additional output types that this version doesn't recognize.
+    #[error(
+        "unknown output file type `{file_name}` in archive \
+         (archive may have been created by a newer version of nextest)"
+    )]
+    UnknownOutputType {
+        /// The file name with the unknown type.
+        file_name: String,
+    },
+
+    /// Archive file exceeds maximum allowed size.
+    #[error(
+        "file `{file_name}` in archive exceeds maximum size ({size} bytes, limit is {limit} bytes)"
+    )]
+    FileTooLarge {
+        /// The file name in the archive.
+        file_name: String,
+
+        /// The size of the file.
+        size: u64,
+
+        /// The maximum allowed size.
+        limit: u64,
+    },
+
+    /// Archive file size doesn't match the header.
+    ///
+    /// This indicates a corrupt or tampered archive since nextest controls
+    /// archive creation.
+    #[error(
+        "file `{file_name}` size mismatch: header claims {claimed_size} bytes, \
+         but read {actual_size} bytes (archive may be corrupt or tampered)"
+    )]
+    SizeMismatch {
+        /// The file name in the archive.
+        file_name: String,
+
+        /// The size claimed in the ZIP header.
+        claimed_size: u64,
+
+        /// The actual size read.
+        actual_size: u64,
+    },
+
+    /// Failed to deserialize metadata.
+    #[error("error deserializing `{file_name}`")]
+    DeserializeMetadata {
+        /// The file name being deserialized.
+        file_name: String,
+
+        /// The underlying error.
+        #[source]
+        error: serde_json::Error,
+    },
+}
+
+/// An error that occurred while reconstructing a TestList from a summary.
+///
+/// Returned by [`TestList::from_summary`](crate::list::TestList::from_summary).
+#[derive(Debug, Error)]
+pub enum TestListFromSummaryError {
+    /// Package not found in the package graph.
+    #[error("package `{name}` (id: `{package_id}`) not found in cargo metadata")]
+    PackageNotFound {
+        /// The package name.
+        name: String,
+
+        /// The package ID that was looked up.
+        package_id: String,
+    },
+
+    /// Error parsing rust build metadata.
+    #[error("error parsing rust build metadata")]
+    RustBuildMeta(#[source] RustBuildMetaParseError),
+}
+
 #[cfg(feature = "self-update")]
 mod self_update_errors {
     use super::*;
+    use crate::update::PrereleaseKind;
     use mukti_metadata::ReleaseStatus;
     use semver::{Version, VersionReq};
 
     /// An error that occurs while performing a self-update.
     ///
     /// Returned by methods in the [`update`](crate::update) module.
-    #[cfg(feature = "self-update")]
     #[derive(Debug, Error)]
     #[non_exhaustive]
     pub enum UpdateError {
@@ -1858,6 +2621,17 @@ mod self_update_errors {
             req: VersionReq,
         },
 
+        /// No stable (non-prerelease) version was found.
+        #[error("no stable version found")]
+        NoStableVersion,
+
+        /// No version matching the requested prerelease kind was found.
+        #[error("no version found matching {} channel", kind.description())]
+        NoVersionForPrereleaseKind {
+            /// The kind of prerelease that was requested.
+            kind: PrereleaseKind,
+        },
+
         /// The specified mukti project was not found.
         #[error("project {not_found} not found in release metadata (known projects: {})", known.join(", "))]
         MuktiProjectNotFound {
@@ -1987,7 +2761,6 @@ mod self_update_errors {
         display_str
     }
 
-    #[cfg(feature = "self-update")]
     /// An error occurred while parsing an [`UpdateVersion`](crate::update::UpdateVersion).
     #[derive(Debug, Error)]
     pub enum UpdateVersionParseError {
@@ -2050,7 +2823,7 @@ mod tests {
         let err2 = StringError::new("err2", Some(err1));
         let err3 = StringError::new("err3\nerr3 line 2", Some(err2));
 
-        insta::assert_snapshot!(format!("{}", DisplayErrorChain::new(&err3)), @r"
+        insta::assert_snapshot!(format!("{}", DisplayErrorChain::new(&err3)), @"
         err3
         err3 line 2
           caused by:
@@ -2076,7 +2849,7 @@ mod tests {
             ErrorList::<StringError>::new("waiting on flowers to bloom", vec![err3.clone()])
                 .expect(">= 1 error");
         insta::assert_snapshot!(format!("{}", error_list), @"err3");
-        insta::assert_snapshot!(format!("{}", DisplayErrorChain::new(&error_list)), @r"
+        insta::assert_snapshot!(format!("{}", DisplayErrorChain::new(&error_list)), @"
         err3
           caused by:
           - err2
@@ -2093,7 +2866,7 @@ mod tests {
         )
         .expect(">= 1 error");
 
-        insta::assert_snapshot!(format!("{}", error_list), @r"
+        insta::assert_snapshot!(format!("{}", error_list), @"
         2 errors occurred waiting for the heat death of the universe:
         * err3
             caused by:
@@ -2105,7 +2878,7 @@ mod tests {
             - err5
             - err4
         ");
-        insta::assert_snapshot!(format!("{}", DisplayErrorChain::new(&error_list)), @r"
+        insta::assert_snapshot!(format!("{}", DisplayErrorChain::new(&error_list)), @"
         2 errors occurred waiting for the heat death of the universe:
         * err3
             caused by:
