@@ -10,14 +10,14 @@ use crate::{
         events::*,
         helpers::{Styles, print_lines_in_chunks},
     },
+    run_mode::NextestRunMode,
 };
 use indicatif::{ProgressBar, ProgressDrawTarget, ProgressStyle};
-use nextest_metadata::RustBinaryId;
+use nextest_metadata::{RustBinaryId, TestCaseName};
 use owo_colors::OwoColorize;
 use std::{
     cmp::{max, min},
     env, fmt,
-    io::IsTerminal,
     str::FromStr,
     time::{Duration, Instant},
 };
@@ -85,7 +85,7 @@ impl fmt::Display for MaxProgressRunning {
 }
 
 /// How to show progress.
-#[derive(Default, Clone, Copy, Debug)]
+#[derive(Default, Clone, Copy, Debug, PartialEq, Eq)]
 pub enum ShowProgress {
     /// Automatically decide based on environment.
     #[default]
@@ -112,7 +112,7 @@ pub(super) enum RunningTestStatus {
 #[derive(Debug)]
 pub(super) struct RunningTest {
     binary_id: RustBinaryId,
-    test_name: String,
+    test_name: TestCaseName,
     status: RunningTestStatus,
     start_time: Instant,
     paused_for: Duration,
@@ -158,6 +158,7 @@ impl RunningTest {
 #[derive(Debug)]
 pub(super) struct ProgressBarState {
     bar: ProgressBar,
+    mode: NextestRunMode,
     stats: RunStats,
     running: usize,
     max_progress_running: MaxProgressRunning,
@@ -187,6 +188,7 @@ pub(super) struct ProgressBarState {
 
 impl ProgressBarState {
     pub(super) fn new(
+        mode: NextestRunMode,
         test_count: usize,
         progress_chars: &str,
         max_progress_running: MaxProgressRunning,
@@ -221,6 +223,7 @@ impl ProgressBarState {
 
         Self {
             bar,
+            mode,
             stats: RunStats::default(),
             running: 0,
             max_progress_running,
@@ -272,7 +275,7 @@ impl ProgressBarState {
     }
 
     fn update_message(&mut self, styles: &Styles) {
-        let mut msg = progress_bar_msg(&self.stats, self.running, styles);
+        let mut msg = self.progress_bar_msg(styles);
         msg += "     ";
 
         if let Some(running_tests) = &self.running_tests {
@@ -289,11 +292,12 @@ impl ProgressBarState {
             }
             if count < running_tests.len() {
                 let overflow_count = running_tests.len() - count;
-                msg.push_str(&format!(
+                swrite!(
+                    msg,
                     "\n             ... and {} more {} running",
                     overflow_count.style(styles.count),
-                    plural::tests_str(overflow_count),
-                ));
+                    plural::tests_str(self.mode, overflow_count),
+                );
                 count += 1;
             }
             self.max_running_displayed = max(self.max_running_displayed, count);
@@ -302,6 +306,10 @@ impl ProgressBarState {
         self.bar.set_message(msg);
     }
 
+    fn progress_bar_msg(&self, styles: &Styles) -> String {
+        progress_bar_msg(&self.stats, self.running, styles)
+    }
+
     pub(super) fn update_progress_bar(&mut self, event: &TestEvent<'_>, styles: &Styles) {
         let before_should_hide = self.should_hide();
 
@@ -333,6 +341,7 @@ impl ProgressBarState {
                 ..
             } => {
                 self.running = *running;
+                self.stats = *current_stats;
 
                 self.bar.set_prefix(progress_bar_prefix(
                     current_stats,
@@ -361,6 +370,7 @@ impl ProgressBarState {
                 ..
             } => {
                 self.running = *running;
+                self.stats = *current_stats;
                 self.remove_test(test_instance);
 
                 self.bar.set_prefix(progress_bar_prefix(
@@ -407,7 +417,7 @@ impl ProgressBarState {
                         .iter_mut()
                         .find(|rt| {
                             &rt.binary_id == test_instance.binary_id
-                                && rt.test_name == test_instance.test_name
+                                && &rt.test_name == test_instance.test_name
                         })
                         .expect("a slow test to be already running")
                         .status = RunningTestStatus::Slow;
@@ -441,8 +451,18 @@ impl ProgressBarState {
                     }
                 }
             }
-            TestEventKind::RunBeginCancel { current_stats, .. }
-            | TestEventKind::RunBeginKill { current_stats, .. } => {
+            TestEventKind::RunBeginCancel {
+                current_stats,
+                running,
+                ..
+            }
+            | TestEventKind::RunBeginKill {
+                current_stats,
+                running,
+                ..
+            } => {
+                self.running = *running;
+                self.stats = *current_stats;
                 self.bar.set_prefix(progress_bar_cancel_prefix(
                     current_stats.cancel_reason,
                     styles,
@@ -467,7 +487,7 @@ impl ProgressBarState {
                     .iter()
                     .position(|e| {
                         &e.binary_id == test_instance.binary_id
-                            && e.test_name == test_instance.test_name
+                            && &e.test_name == test_instance.test_name
                     })
                     .expect("finished test to have started"),
             );
@@ -501,16 +521,22 @@ impl ProgressBarState {
     }
 }
 
-/// OSC 9 terminal progress reporting.
-#[derive(Default)]
-pub(super) struct TerminalProgress {
-    last_value: TerminalProgressValue,
+/// Whether to show OSC 9;4 terminal progress.
+#[derive(Clone, Copy, Debug, Eq, PartialEq)]
+pub enum ShowTerminalProgress {
+    /// Show terminal progress.
+    Yes,
+
+    /// Do not show terminal progress.
+    No,
 }
 
-impl TerminalProgress {
+impl ShowTerminalProgress {
     const ENV: &str = "CARGO_TERM_PROGRESS_TERM_INTEGRATION";
 
-    pub(super) fn new(configs: &CargoConfigs, stream: &dyn IsTerminal) -> Option<Self> {
+    /// Determines whether to show terminal progress based on Cargo configs and
+    /// whether the output is a terminal.
+    pub fn from_cargo_configs(configs: &CargoConfigs, is_terminal: bool) -> Self {
         // See whether terminal integration is enabled in Cargo.
         for config in configs.discovered_configs() {
             match config {
@@ -518,10 +544,10 @@ impl TerminalProgress {
                     if let Some(v) = config.term.progress.term_integration {
                         if v {
                             debug!("enabling terminal progress reporting based on {source:?}");
-                            return Some(Self::default());
+                            return Self::Yes;
                         } else {
                             debug!("disabling terminal progress reporting based on {source:?}");
-                            return None;
+                            return Self::No;
                         }
                     }
                 }
@@ -532,13 +558,13 @@ impl TerminalProgress {
                                 "enabling terminal progress reporting based on \
                                  CARGO_TERM_PROGRESS_TERM_INTEGRATION environment variable"
                             );
-                            return Some(Self::default());
+                            return Self::Yes;
                         } else if v == "false" {
                             debug!(
                                 "disabling terminal progress reporting based on \
                                  CARGO_TERM_PROGRESS_TERM_INTEGRATION environment variable"
                             );
-                            return None;
+                            return Self::No;
                         } else {
                             debug!(
                                 "invalid value for CARGO_TERM_PROGRESS_TERM_INTEGRATION \
@@ -551,17 +577,36 @@ impl TerminalProgress {
                     if let Some(v) = config.term.progress.term_integration {
                         if v {
                             debug!("enabling terminal progress reporting based on {source:?}");
-                            return Some(Self::default());
+                            return Self::Yes;
                         } else {
                             debug!("disabling terminal progress reporting based on {source:?}");
-                            return None;
+                            return Self::No;
                         }
                     }
                 }
             }
         }
 
-        supports_osc_9_4(stream).then(Self::default)
+        if supports_osc_9_4(is_terminal) {
+            Self::Yes
+        } else {
+            Self::No
+        }
+    }
+}
+
+/// OSC 9 terminal progress reporting.
+#[derive(Default)]
+pub(super) struct TerminalProgress {
+    last_value: TerminalProgressValue,
+}
+
+impl TerminalProgress {
+    pub(super) fn new(show: ShowTerminalProgress) -> Option<Self> {
+        match show {
+            ShowTerminalProgress::Yes => Some(Self::default()),
+            ShowTerminalProgress::No => None,
+        }
     }
 
     pub(super) fn update_progress(&mut self, event: &TestEvent<'_>) {
@@ -617,8 +662,8 @@ impl TerminalProgress {
 }
 
 /// Determines whether the terminal supports ANSI OSC 9;4.
-fn supports_osc_9_4(stream: &dyn IsTerminal) -> bool {
-    if !stream.is_terminal() {
+fn supports_osc_9_4(is_terminal: bool) -> bool {
+    if !is_terminal {
         debug!(
             "autodetect terminal progress reporting: disabling since \
              passed-in stream (usually stderr) is not a terminal"
@@ -634,7 +679,7 @@ fn supports_osc_9_4(stream: &dyn IsTerminal) -> bool {
         return true;
     }
     if let Ok(term) = std::env::var("TERM_PROGRAM")
-        && (term == "WezTerm" || term == "ghostty")
+        && (term == "WezTerm" || term == "ghostty" || term == "iTerm.app")
     {
         debug!("autodetect terminal progress reporting: enabling since TERM_PROGRAM is {term}");
         return true;
@@ -856,6 +901,12 @@ pub(super) fn progress_bar_msg(
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::{
+        reporter::TestOutputDisplay,
+        test_output::{ChildExecutionOutput, ChildOutput, ChildSingleOutput, ChildSplitOutput},
+    };
+    use bytes::Bytes;
+    use chrono::Local;
 
     #[test]
     fn test_progress_bar_prefix() {
@@ -947,7 +998,7 @@ mod tests {
 
     fn running_test_examples(now: Instant) -> Vec<(&'static str, RunningTest)> {
         let binary_id = RustBinaryId::new("my-binary");
-        let test_name = "test::my_test".to_string();
+        let test_name = TestCaseName::new("test::my_test");
         let start_time = now - Duration::from_secs(125); // 2 minutes 5 seconds ago
 
         vec![
@@ -1091,4 +1142,202 @@ mod tests {
             ),
         ]
     }
+
+    /// Test that `update_progress_bar` correctly updates `self.stats` when
+    /// processing `TestStarted` and `TestFinished` events.
+    ///
+    /// This test verifies both:
+    ///
+    /// 1. State: `self.stats` equals the event's `current_stats` after processing.
+    /// 2. Output: `state.progress_bar_msg()` reflects the updated stats.
+    #[test]
+    fn update_progress_bar_updates_stats() {
+        let styles = Styles::default();
+        let binary_id = RustBinaryId::new("test-binary");
+        let test_name = TestCaseName::new("test_name");
+
+        // Create ProgressBarState with initial (default) stats.
+        let mut state = ProgressBarState::new(
+            NextestRunMode::Test,
+            10,
+            "=> ",
+            MaxProgressRunning::default(),
+        );
+
+        // Verify the initial state.
+        assert_eq!(state.stats, RunStats::default());
+        assert_eq!(state.running, 0);
+
+        // Create a TestStarted event.
+        let started_stats = RunStats {
+            initial_run_count: 10,
+            passed: 5,
+            ..RunStats::default()
+        };
+        let started_event = TestEvent {
+            timestamp: Local::now().fixed_offset(),
+            elapsed: Duration::ZERO,
+            kind: TestEventKind::TestStarted {
+                stress_index: None,
+                test_instance: TestInstanceId {
+                    binary_id: &binary_id,
+                    test_name: &test_name,
+                },
+                current_stats: started_stats,
+                running: 3,
+                command_line: vec![],
+            },
+        };
+
+        state.update_progress_bar(&started_event, &styles);
+
+        // Verify the state was updated.
+        assert_eq!(
+            state.stats, started_stats,
+            "stats should be updated on TestStarted"
+        );
+        assert_eq!(state.running, 3, "running should be updated on TestStarted");
+
+        // Verify that the output reflects the updated stats.
+        let msg = state.progress_bar_msg(&styles);
+        insta::assert_snapshot!(msg, @"3 running, 5 passed, 0 skipped");
+
+        // Create a TestFinished event with different stats.
+        let finished_stats = RunStats {
+            initial_run_count: 10,
+            finished_count: 1,
+            passed: 8,
+            ..RunStats::default()
+        };
+        let finished_event = TestEvent {
+            timestamp: Local::now().fixed_offset(),
+            elapsed: Duration::ZERO,
+            kind: TestEventKind::TestFinished {
+                stress_index: None,
+                test_instance: TestInstanceId {
+                    binary_id: &binary_id,
+                    test_name: &test_name,
+                },
+                success_output: TestOutputDisplay::Never,
+                failure_output: TestOutputDisplay::Never,
+                junit_store_success_output: false,
+                junit_store_failure_output: false,
+                run_statuses: ExecutionStatuses::new(vec![ExecuteStatus {
+                    retry_data: RetryData {
+                        attempt: 1,
+                        total_attempts: 1,
+                    },
+                    output: make_test_output(),
+                    result: ExecutionResultDescription::Pass,
+                    start_time: Local::now().fixed_offset(),
+                    time_taken: Duration::from_secs(1),
+                    is_slow: false,
+                    delay_before_start: Duration::ZERO,
+                    error_summary: None,
+                    output_error_slice: None,
+                }]),
+                current_stats: finished_stats,
+                running: 2,
+            },
+        };
+
+        state.update_progress_bar(&finished_event, &styles);
+
+        // Verify the state was updated.
+        assert_eq!(
+            state.stats, finished_stats,
+            "stats should be updated on TestFinished"
+        );
+        assert_eq!(
+            state.running, 2,
+            "running should be updated on TestFinished"
+        );
+
+        // Verify that the output reflects the updated stats.
+        let msg = state.progress_bar_msg(&styles);
+        insta::assert_snapshot!(msg, @"2 running, 8 passed, 0 skipped");
+
+        // Create a RunBeginCancel event.
+        let cancel_stats = RunStats {
+            initial_run_count: 10,
+            finished_count: 3,
+            passed: 2,
+            failed: 1,
+            cancel_reason: Some(CancelReason::TestFailure),
+            ..RunStats::default()
+        };
+        let cancel_event = TestEvent {
+            timestamp: Local::now().fixed_offset(),
+            elapsed: Duration::ZERO,
+            kind: TestEventKind::RunBeginCancel {
+                setup_scripts_running: 0,
+                current_stats: cancel_stats,
+                running: 4,
+            },
+        };
+
+        state.update_progress_bar(&cancel_event, &styles);
+
+        // Verify the state was updated.
+        assert_eq!(
+            state.stats, cancel_stats,
+            "stats should be updated on RunBeginCancel"
+        );
+        assert_eq!(
+            state.running, 4,
+            "running should be updated on RunBeginCancel"
+        );
+
+        // Verify that the output reflects the updated stats.
+        let msg = state.progress_bar_msg(&styles);
+        insta::assert_snapshot!(msg, @"4 running, 2 passed, 1 failed, 0 skipped");
+
+        // Create a RunBeginKill event with different stats.
+        let kill_stats = RunStats {
+            initial_run_count: 10,
+            finished_count: 5,
+            passed: 3,
+            failed: 2,
+            cancel_reason: Some(CancelReason::Signal),
+            ..RunStats::default()
+        };
+        let kill_event = TestEvent {
+            timestamp: Local::now().fixed_offset(),
+            elapsed: Duration::ZERO,
+            kind: TestEventKind::RunBeginKill {
+                setup_scripts_running: 0,
+                current_stats: kill_stats,
+                running: 2,
+            },
+        };
+
+        state.update_progress_bar(&kill_event, &styles);
+
+        // Verify the state was updated.
+        assert_eq!(
+            state.stats, kill_stats,
+            "stats should be updated on RunBeginKill"
+        );
+        assert_eq!(
+            state.running, 2,
+            "running should be updated on RunBeginKill"
+        );
+
+        // Verify that the output reflects the updated stats.
+        let msg = state.progress_bar_msg(&styles);
+        insta::assert_snapshot!(msg, @"2 running, 3 passed, 2 failed, 0 skipped");
+    }
+
+    // Helper to create minimal output for ExecuteStatus.
+    fn make_test_output() -> ChildExecutionOutputDescription<ChildSingleOutput> {
+        ChildExecutionOutput::Output {
+            result: Some(ExecutionResult::Pass),
+            output: ChildOutput::Split(ChildSplitOutput {
+                stdout: Some(Bytes::new().into()),
+                stderr: Some(Bytes::new().into()),
+            }),
+            errors: None,
+        }
+        .into()
+    }
 }
